
#include <linux/module.h>
#include <linux/init.h>
#include <linux/dmaengine.h>
#include <linux/dma-mapping.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/completion.h>



#define SRC_BUF_SIZE 64  // Size of the source buffer
#define DMA_TRANSFER_SIZE 64  // Size of each transfer in bytes
#define DEST_ADDR	0x20400000U

static struct dma_chan *dma_chan;
static dma_addr_t src_dma_addr, dst_dma_addr;
static char *src_buf, *dst_buf;
static struct completion dma_complete;

static void dma_callback(void *completion)
{
    complete(completion);  // Signal the completion of the DMA transfer
}

static int edma_transfer_test(void)
{
    struct dma_async_tx_descriptor *tx;
    struct dma_slave_config config;
    enum dma_ctrl_flags flags = DMA_CTRL_ACK | DMA_PREP_INTERRUPT;
    struct dma_device *dma_dev;
    dma_cookie_t cookie;
    int ret;

    // Step 1: Allocate DMA-capable memory for source and destination buffers
    src_buf = kzalloc(SRC_BUF_SIZE, GFP_KERNEL);
    if (!src_buf)
        return -ENOMEM;

/*    dst_buf = kzalloc(SRC_BUF_SIZE, GFP_KERNEL);
    if (!dst_buf) {
        kfree(src_buf);
        return -ENOMEM;
    } */

//    memset(src_dma_addr, 0xAA, SRC_BUF_SIZE);
    memset(src_buf, 0xAA, SRC_BUF_SIZE);

    // Step 2: Map the buffers to DMA addresses
    src_dma_addr = dma_map_single(dma_chan->device->dev, src_buf, DMA_TRANSFER_SIZE, DMA_TO_DEVICE);
//    dst_dma_addr = dma_map_single(dma_chan->device->dev, dst_buf, DMA_TRANSFER_SIZE, DMA_FROM_DEVICE);
/*
    if (dma_mapping_error(dma_chan->device->dev, src_dma_addr) ||
        dma_mapping_error(dma_chan->device->dev, dst_dma_addr)) {
        kfree(src_buf);
        kfree(dst_buf);
        return -ENOMEM;
    }
*/

    // Step 3: Configure the DMA transfer (source and destination addresses, burst sizes)
    memset(&config, 0, sizeof(config));
    config.src_addr = src_buf;
    config.dst_addr = DEST_ADDR;
    config.src_addr_width = DMA_SLAVE_BUSWIDTH_4_BYTES;
    config.dst_addr_width = DMA_SLAVE_BUSWIDTH_4_BYTES;
    config.src_maxburst = 16;
    config.dst_maxburst = 16;

    ret = dmaengine_slave_config(dma_chan, &config);
    if (ret) {
        pr_err("Failed to configure DMA channel\n");
        goto unmap_dma;
    }

    // Step 4: Prepare the DMA transaction
    tx = dmaengine_prep_dma_memcpy(dma_chan, dst_dma_addr, src_dma_addr, DMA_TRANSFER_SIZE, flags);
    if (!tx) {
        pr_err("Failed to prepare DMA transaction\n");
        ret = -EINVAL;
        goto unmap_dma;
    }

    // Set up a completion callback for the DMA transfer
    tx->callback = dma_callback;
    tx->callback_param = &dma_complete;

    // Step 5: Submit the DMA transaction and start the transfer
    init_completion(&dma_complete);
    cookie = dmaengine_submit(tx);
    if (dma_submit_error(cookie)) {
        pr_err("Failed to submit DMA transaction\n");
        ret = -EINVAL;
        goto unmap_dma;
    }

    // Start the DMA transfer
    dma_async_issue_pending(dma_chan);

    // Wait for the DMA transfer to complete
    wait_for_completion(&dma_complete);

    pr_info("DMA transfer completed successfully\n");

unmap_dma:
    dma_unmap_single(dma_chan->device->dev, src_dma_addr, DMA_TRANSFER_SIZE, DMA_TO_DEVICE);
//    dma_unmap_single(dma_chan->device->dev, dst_dma_addr, DMA_TRANSFER_SIZE, DMA_FROM_DEVICE);

    kfree(src_buf);
    kfree(dst_buf);

    return ret;
}

static int __init edma_test_init(void)
{
    pr_info("edma_test_init\n");
    // Step 1: Request a DMA channel
 
    dma_chan = dma_request_chan_by_mask((dma_cap_mask_t *)DMA_SLAVE);
    pr_info("DMA request chan\n");
    if (IS_ERR(dma_chan)) {
        pr_err("Failed to request DMA channel\n");
        return PTR_ERR(dma_chan);
    }

    pr_info("EDMA transfer test module loaded\n");
    return edma_transfer_test();
}

static void __exit edma_test_exit(void)
{
    if (dma_chan)
        dma_release_channel(dma_chan);

    pr_info("EDMA transfer test module unloaded\n");
}

module_init(edma_test_init);
module_exit(edma_test_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Your Name");
MODULE_DESCRIPTION("EDMA Transfer Test on AM5748");

